import inspect
from collections.abc import Callable

from mafunca.common.exceptions import MonadError, CurryBadFunctionError


def extract_name(func) -> str:
    return getattr(func, "__qualname__", getattr(func, "__name__", f"{func}"))


def on_coroutine(fn: Callable, monad_name: str, method: str):
    """
       Panic when the monadic contract is violated - function must be sync.
       :raises MonadError:
    """
    if inspect.iscoroutinefunction(fn):
        raise MonadError(
            monad_name,
            method,
            message=f"function '{extract_name(fn)}' - async function can not be used"
        )


def on_sync(fn: Callable, monad_name: str, method: str):
    """
       Panic when the monadic contract is violated - function must be async.
       :raises MonadError:
    """
    if not inspect.iscoroutinefunction(fn):
        raise MonadError(
            monad_name,
            method,
            message=f"function '{extract_name(fn)}' - sync function can not be used")


def on_monadic_result(result, fn: Callable, monad, method: str):
    """
       Panic when the monadic contract is violated - the function should not return a monad.
       :raises MonadError:
    """
    if isinstance(result, monad):
        name = monad.__name__
        raise MonadError(
            name,
            method,
            f"return value {result} of applying function '{extract_name(fn)}' - must not be '{name}' entity"
        )


def on_another_instance(result, fn: Callable, monad, method: str):
    """
       Panic when the monadic contract is violated - the function should return the same instance.
       :raises MonadError:
    """
    if not isinstance(result, monad):
        name = monad.__name__
        raise MonadError(
            name,
            method,
            f"return value {result} of applying function '{extract_name(fn)}' must be '{name}' entity"
        )


def on_bad_curried(func):
    """
       Panic on improper entity for currying.
       :raises CurryBadFunctionError:
    """
    if not inspect.isfunction(func):
        raise CurryBadFunctionError(func_name=extract_name(func), err="must be a callable")
    if inspect.isbuiltin(func):
        raise CurryBadFunctionError(func_name=extract_name(func), err="should not be a built-in function")
    if inspect.ismethod(func):
        raise CurryBadFunctionError(func_name=extract_name(func), err="should not be a bound method")


def curry_on_coroutine(func: Callable):
    """
       Panic on async function in sync curry decorator.
       :raises CurryBadFunctionError:
    """
    if inspect.iscoroutinefunction(func):
        raise CurryBadFunctionError(
            func_name=extract_name(func),
            err="should not be an async function. Use async curry decorator instead."
        )


def curry_on_sync(func: Callable):
    """
       Panic on sync function in async curry decorator.
       :raises CurryBadFunctionError:
    """
    if not inspect.iscoroutinefunction(func):
        raise CurryBadFunctionError(
            func_name=extract_name(func),
            err="should not be a sync function. Use sync curry decorator instead."
        )



import inspect
from collections.abc import Callable, Awaitable
from typing import TypeVar, ParamSpec, Generic, Union, List, Dict

from mafunca.common.exceptions import CurryBadArguments
import mafunca.common.panics as panics


__all__ = ['curry', 'async_curry', 'Curry', 'AsyncCurry']


def _in_place_endpoints_filter(sig: inspect.Signature, bound_args: inspect.BoundArguments) -> inspect.BoundArguments:
    """Cleaning arguments of the *args and **kwargs type - they should not have default values applied to them"""
    for name, par in sig.parameters.items():
        if par.kind == inspect.Parameter.VAR_KEYWORD or par.kind == inspect.Parameter.VAR_POSITIONAL:
            bound_args.arguments.pop(name, None)
    return bound_args


def _apply(sig: inspect.Signature, *args, **kwargs) -> inspect.BoundArguments:
    """
        Applying arguments to a function signature.
        :raises TypeError: error of 'bind_partial' method.
    """
    bound_args = sig.bind_partial(*args, **kwargs)
    if len(args) == 0 and len(kwargs) == 0:
        bound_args.apply_defaults()
        _in_place_endpoints_filter(sig=sig, bound_args=bound_args)
    return bound_args


A = ParamSpec("A")
R = TypeVar("R")


class _Base:
    __slots__ = ('_func', '_sig', '_pos', '_named')

    def __init__(self, fn: Callable[[A], R]):
        self._func = fn
        self._sig = inspect.signature(fn)
        self._pos = []
        self._named = {}

    def _update_state(self, sig: inspect.Signature, pos: List, named: Dict):
        self._sig = sig
        self._pos = pos
        self._named = named

    @property
    def origin(self) -> Callable[[A], R]:
        return self._func

    def __repr__(self):
        cls = type(self)
        name = cls.__qualname__
        module = cls.__module__
        args = [repr(self._func)]
        args.extend(repr(x) for x in self._pos)
        args.extend(f"{k}={v!r}" for (k, v) in self._named.items())
        return f"{module}.{name}({', '.join(args)})"


class Curry(Generic[R], _Base):
    def __init__(self, fn: Callable[[A], R]):
        super().__init__(fn=fn)

    def __call__(self, *args: A.args, **kwargs: A.kwargs) -> Union['Curry[R]', R]:
        try:
            bound_args = _apply(self._sig, *args, **kwargs)
            new_params = [par for name, par in self._sig.parameters.items() if name not in bound_args.arguments]
            if len(new_params) == 0:
                return self._func(*self._pos, *bound_args.args, **self._named, **bound_args.kwargs)

            sig = inspect.Signature(parameters=new_params)
            new_curry = Curry(self._func)
            new_curry._update_state(sig, [*self._pos, *bound_args.args], {**self._named, **bound_args.kwargs})
            return new_curry
        except TypeError as err:
            raise CurryBadArguments(func_name=panics.extract_name(self._func), err=err.args[0]) from None

    def run_for_var(self) -> R:
        """
           Runs a function whose signature contains variable arguments without waiting for them to be passed.
           IMPORTANT: make sure that only variable arguments are left unset.
        """
        return self._func(*self._pos, **self._named)


class AsyncCurry(Generic[R], _Base):
    def __init__(self, fn: Callable[[A], Awaitable[R]]):
        super().__init__(fn=fn)

    async def __call__(self, *args: A.args, **kwargs: A.kwargs) -> Union['AsyncCurry[R]', R]:
        try:
            bound_args = _apply(self._sig, *args, **kwargs)
            new_params = [par for name, par in self._sig.parameters.items() if name not in bound_args.arguments]
            if len(new_params) == 0:
                return await self._func(*self._pos, *bound_args.args, **self._named, **bound_args.kwargs)

            sig = inspect.Signature(parameters=new_params)
            new_curry = AsyncCurry(self._func)
            new_curry._update_state(sig, [*self._pos, *bound_args.args], {**self._named, **bound_args.kwargs})
            return new_curry
        except TypeError as err:
            raise CurryBadArguments(func_name=panics.extract_name(self._func), err=err.args[0]) from None

    async def run_for_var(self) -> R:
        """
           Runs a function whose signature contains variable arguments without waiting for them to be passed.
           IMPORTANT: make sure that only variable arguments are left unset.
        """
        return await self._func(*self._pos, **self._named)


def curry(fn: Callable[[A], R]) -> Curry[R]:
    """
       Decorator that turns a SYNC ONLY function into a curried version.
       :raises CurryBadFunctionError: passed function is not suitable
       :raises CurryBadArguments: error at the level of the arguments being passed
    """
    panics.on_bad_curried(func=fn)
    panics.curry_on_coroutine(func=fn)
    curried = Curry(fn)
    curried.__doc__ = fn.__doc__
    return curried


def async_curry(fn: Callable[[A], Awaitable[R]]) -> AsyncCurry[R]:
    """
       Decorator that turns an ASYNC function into a curried version.
       :raises CurryBadFunctionError: passed function is not suitable
       :raises CurryBadArguments: error at the level of the arguments being passed
    """
    panics.on_bad_curried(func=fn)
    panics.curry_on_sync(func=fn)
    curried = AsyncCurry(fn)
    curried.__doc__ = fn.__doc__
    return curried



import unittest
import asyncio

from mafunca.triple import Right, Left, Nothing, TUtils, impure
from mafunca.common.exceptions import MonadError, CurryBadArguments
from mafunca.curry import curry, async_curry


class TestTripleAndCurry(unittest.TestCase):
    def test_introspection_right_left_nothing(self):
        right = Right(1)
        self.assertTrue(right.is_right)
        self.assertFalse(right.is_nothing)

        from_unit = TUtils.unit(1)
        self.assertTrue(from_unit.is_right)
        self.assertFalse(from_unit.is_nothing)

        left = Left(1)
        self.assertFalse(left.is_right)
        self.assertFalse(left.is_nothing)

        nothing = Nothing()
        self.assertFalse(nothing.is_right)
        self.assertTrue(nothing.is_nothing)

        self.assertFalse(TUtils.is_bad(Right(1)))
        self.assertTrue(TUtils.is_bad(Left(0)))
        self.assertTrue(TUtils.is_bad(Nothing()))

    def test_chain_right(self):
        res = TUtils.unit(2).map(lambda x: x + 1).bind(lambda x: Right(x + 1)).unfold(right=lambda x: x ** 2)
        self.assertEqual(res, 16)

        res = Right(1).map(lambda x: x + 1).bind(lambda x: Right(x + 1)).get_or_else(100)
        self.assertEqual(res, 3)

        res = (
            Right(0)
            .map(lambda x: x + 1)
            .recover_from_left(lambda _: 100)
            .bind(lambda x: TUtils.unit(x + 1))
            .map(lambda x: x + 1)
            .recover_from_nothing(lambda: 1000)
            .bind(lambda x: Right(x + 1))
            .unfold(right=lambda x: x ** 2)
        )
        self.assertEqual(res, 16)

    def test_chain_left(self):
        res = Left(1).map(lambda x: x + 1).bind(lambda x: Right(x + 1))
        self.assertIsInstance(res, Left)

        res = Left(1).map(lambda x: x + 1).bind(lambda x: Right(x + 1)).get_or_else(100)
        self.assertEqual(res, 100)

        res = Left("Some error").map(lambda x: x + 1).bind(lambda x: Right(x + 1)).unfold(left=lambda x: [x])
        self.assertEqual(res, ["Some error"])

        res = (
            Left("Some error")
            .map(lambda x: x + 1)
            .recover_from_left(lambda _: 100)
            .recover_from_nothing(lambda: 1000)
            .recover_from_left(lambda _: 200)
            .unfold(right=lambda x: [x])
        )
        self.assertEqual(res, [100])

    def test_chain_nothing(self):
        res = Nothing().map(lambda x: x + 1).bind(lambda x: Right(x + 1))
        self.assertIsInstance(res, Nothing)

        res = Nothing().map(lambda x: x + 1).bind(lambda x: Right(x + 1)).get_or_else(0)
        self.assertEqual(res, 0)

        res = (
            Nothing()
            .map(lambda x: x + 1)
            .recover_from_left(lambda _: 100)
            .recover_from_nothing(lambda: 1000)
            .bind(lambda x: Right(x + 1))
            .recover_from_nothing(lambda: 1)
            .unfold()
        )
        self.assertEqual(res, 1001)

    def test_chains_violations(self):
        def fn_monadic(x):
            return Right(x)

        async def fn_async(x):
            return x

        @impure
        def fn_impure(a):
            return a

        with self.assertRaises(MonadError):
            Right(1).map(fn_monadic)
        with self.assertRaises(MonadError):
            Right(1).map(fn_async)
        with self.assertRaises(MonadError):
            Right(1).map(fn_impure)

    def test_from_nullable(self):
        self.assertIsInstance(TUtils.from_nullable(None), Nothing)
        self.assertIsInstance(TUtils.from_nullable({"a": 1}, predicate=lambda d: d.get("b")), Nothing)
        self.assertIsInstance(TUtils.from_nullable({"a": 1}, predicate=lambda d: d.get("a")), Right)
        self.assertIsInstance(TUtils.from_nullable(1, predicate=lambda _: False), Nothing)
        self.assertIsInstance(TUtils.from_nullable(None, predicate=lambda a: a is None), Right)

    def test_from_try(self):
        @TUtils.from_try
        def test_from_try(a):
            return a ** 2

        res = test_from_try("1").map(lambda x: x + 1).bind(lambda x: Right(x))
        self.assertTrue(TUtils.is_bad(res))
        self.assertTrue(res.unfold(left=lambda e: isinstance(e, TypeError)))

        res = test_from_try(10).map(lambda x: x + 1).bind(lambda x: Right(x + 1))
        self.assertFalse(TUtils.is_bad(res))
        self.assertEqual(res.unfold(right=lambda x: x + 1), 103)

    def test_closer(self):
        @TUtils.closer
        def test_closer(a: int, b: int, c: int) -> int:
            return a + b + c

        self.assertIsInstance(test_closer(Left(1), 2, 3), Left)
        self.assertIsInstance(test_closer(1, 2, Nothing()), Nothing)
        self.assertIsInstance(test_closer(1, Left(2), Nothing()), Left)
        self.assertEqual(test_closer(Right(1), Right(2), Right(3)), 6)
        self.assertEqual(test_closer(1, 2, 3), 6)
        self.assertEqual(test_closer(1, Right(2), 3), 6)

    def test_curry_basic(self):
        @curry
        def for_curry_one(a: int, b: int, c: int = 0, d: int = 0) -> list[int]:
            return [a, b, c, d]

        @curry
        def for_curry_two(a: int, b: int, *args, **kwargs) -> list:
            return [a, b, args, kwargs]

        @curry
        def for_curry_three(a: int, *, b: int) -> int:
            return a + b

        self.assertEqual(for_curry_one(1)(2)(3)(4), [1, 2, 3, 4])
        self.assertEqual(for_curry_one(1)(2)(), [1, 2, 0, 0])
        self.assertEqual(for_curry_one()(b=2, a=1), [1, 2, 0, 0])
        self.assertEqual(for_curry_one(1, 2)(c=3)(), [1, 2, 3, 0])
        self.assertEqual(for_curry_one(1, 2)(d=3)(), [1, 2, 0, 3])

        res = for_curry_two(1, b=2)
        self.assertTrue(callable(for_curry_two))
        res = res(0, 0)
        self.assertTrue(callable(for_curry_two))
        res = res(another=10)
        self.assertEqual(res, [1, 2, (0, 0), {'another': 10}])
        res = for_curry_two(b=2)(a=1)
        self.assertEqual(res.run_for_var(), [1, 2, (), {}])

        with self.assertRaises(CurryBadArguments):
            for_curry_three(1)(2)  # signature of the original function are preserved - the second arg is only named
        self.assertEqual(for_curry_three(1)(b=2), 3)

    def test_lift(self):
        @curry
        def for_curry(a: int, b: int, c: int = 0, d: int = 0) -> list[int]:
            return [a, b, c, d]

        res = TUtils.lift(for_curry, Right(1), Right(2), Right(3), Right(4))
        self.assertEqual(res.unfold(), [1, 2, 3, 4])

        res = TUtils.lift(for_curry, Right(1), Left('some error'), Right(3), Right(4))
        self.assertEqual(res.unfold(), 'some error')


class TestAsyncCurry(unittest.IsolatedAsyncioTestCase):
    async def test_async_curry(self):
        @async_curry
        async def for_curry(a: int, b: int, c: int = 0, d: int = 0) -> list[int]:
            await asyncio.sleep(0)
            return [a, b, c, d]

        res = await for_curry(1)
        res = await res(2)
        res = await res(3)
        res = await res(4)
        self.assertEqual(res, [1, 2, 3, 4])

        res = await for_curry()
        res = await res(b=2, a=1)
        self.assertEqual(res, [1, 2, 0, 0])


if __name__ == "__main__":
    unittest.main()



class MonadError(Exception):
    """Thrown when monadic contracts are violated.
       Not recommended to import and catch.
       Separated - not in the error hierarchy of this library.
    """
    def __init__(self, monad: str, method: str, message: str):
        text = f"violation of the {monad}-{method} contract - {message}"
        super().__init__(text)


class BaseLibError(Exception):
    """Base library level exception"""
    def __init__(self, message: str):
        super().__init__(message)


class ImpureMarkError(BaseLibError):
    """Thrown when an attempt is made to mark a function as impure"""
    def __init__(self, func_name: str, err: str):
        text = f"Can't mark this function {func_name} as impure: {err}"
        super().__init__(text)


class CurryBadFunctionError(BaseLibError):
    """Thrown when the 'curried' function is not suitable"""
    def __init__(self, func_name: str, err: str):
        text = f"{func_name} - {err}"
        super().__init__(text)


class CurryBadArguments(BaseLibError):
    """thrown when the passed arguments for the function are incorrect"""
    def __init__(self, func_name: str, err: str):
        text = f"{func_name} - {err}"
        super().__init__(text)



from typing import TypeVar, Generic, overload, Union, Optional
from collections.abc import Callable, Awaitable
import inspect
import asyncio

from mafunca.triple import Left, Nothing, TUtils
from mafunca.common.exceptions import MonadError
import mafunca.common.panics as panics


__all__ = ['Eff']


A = TypeVar('A')
B = TypeVar('B')
C = TypeVar('C')

L = TypeVar('L')
Exc = TypeVar('Exc', bound=Exception)
T = TypeVar('T')


async def _maybe_await(obj: Union[Awaitable[T], T]) -> T:
    return await obj if inspect.isawaitable(obj) else obj


class Eff(Generic[A]):
    """Lazy monad for async effects.
       It can work with bad 'Triple' entities using the short-circuit principle.
       Can work with both synchronous and asynchronous functions.
    """

    __slots__ = ["effect"]

    def __init__(self, effect: Callable[[], A]):
        self.effect = effect

    @overload
    def map(self: 'Eff[Left[L]]', fn: Callable[[B], C]) -> 'Eff[Left[L]]': pass
    @overload
    def map(self: 'Eff[Nothing]', fn: Callable[[B], C]) -> 'Eff[Nothing]': pass
    @overload
    def map(self, fn: Callable[[A], Awaitable[B]]) -> 'Eff[B]': pass
    @overload
    def map(self, fn: Callable[[A], B]) -> 'Eff[B]': pass

    def map(self, fn):
        """
           Applies a function that returns a non-Eff entity.
           It can accept both async and sync functions. Async functions are awaited.
           :raises MonadError: violation of the contract
        """

        async def new_effect():
            previous = await _maybe_await(self.effect())
            if TUtils.is_bad(previous):
                return previous
            current = await _maybe_await(fn(previous))
            panics.on_monadic_result(current, fn=fn, monad=self.__class__, method='map')
            return current
        return Eff(new_effect)

    @overload
    def map_to_thread(self: 'Eff[Left[L]]', fn: Callable[[B], C]) -> 'Eff[Left[L]]': pass
    @overload
    def map_to_thread(self: 'Eff[Nothing]', fn: Callable[[B], C]) -> 'Eff[Nothing]': pass
    @overload
    def map_to_thread(self, fn: Callable[[A], B]) -> 'Eff[B]': pass

    def map_to_thread(self, fn):
        """
           Applies a SYNC ONLY function that returns a non-Eff entity.
           Executes it in a separate thread.
           :raises MonadError: violation of the contract
        """
        panics.on_coroutine(fn, monad_name=self.__class__.__name__, method='map_to_thread')

        async def new_effect():
            previous = await _maybe_await(self.effect())
            if TUtils.is_bad(previous):
                return previous
            current = await asyncio.to_thread(fn, previous)
            panics.on_monadic_result(current, fn=fn, monad=self.__class__, method='map_to_thread')
            return current

        return Eff(new_effect)

    @overload
    def bind(self: 'Eff[Left[L]]', fn: Callable[[B], C]) -> 'Eff[Left[L]]': pass
    @overload
    def bind(self: 'Eff[Nothing]', fn: Callable[[B], C]) -> 'Eff[Nothing]': pass
    @overload
    def bind(self, fn: Callable[[A], Awaitable['Eff[B]']]) -> 'Eff[B]': pass
    @overload
    def bind(self, fn: Callable[[A], 'Eff[B]']) -> 'Eff[B]': pass

    def bind(self, fn):
        """
           Applies a function that returns an Eff entity.
           It can accept both async and sync functions. Async functions are awaited.
           :raises MonadError: violation of the contract
        """

        async def new_effect():
            previous = await _maybe_await(self.effect())
            if TUtils.is_bad(previous):
                return previous
            current_effect = await _maybe_await(fn(previous))
            panics.on_another_instance(current_effect, fn=fn, monad=self.__class__, method='bind')
            current = await _maybe_await(current_effect.effect())
            return current
        return Eff(new_effect)

    @overload
    def bind_to_thread(self: 'Eff[Left[L]]', fn: Callable[[B], C]) -> 'Eff[Left[L]]': pass
    @overload
    def bind_to_thread(self: 'Eff[Nothing]', fn: Callable[[B], C]) -> 'Eff[Nothing]': pass
    @overload
    def bind_to_thread(self, fn: Callable[[A], 'Eff[B]']) -> 'Eff[B]': pass

    def bind_to_thread(self, fn):
        """
           Applies a SYNC ONLY function that returns an Eff entity.
           Executes it in a separate thread - ONLY inner function inside Eff.
           :raises MonadError: violation of the contract
        """

        async def new_effect():
            previous = await _maybe_await(self.effect())
            if TUtils.is_bad(previous):
                return previous
            current_effect = await _maybe_await(fn(previous))
            panics.on_another_instance(current_effect, fn=fn, monad=self.__class__, method='bind_to_thread')
            panics.on_coroutine(current_effect.effect, monad_name=self.__class__.__name__, method='bind_to_thread')
            current = await asyncio.to_thread(current_effect.effect)
            return current

        return Eff(new_effect)

    @overload
    def catch(self, fn: Callable[[Exc], Awaitable['Eff[B]']]) -> 'Eff[B]': pass
    @overload
    def catch(self, fn: Callable[[Exc], 'Eff[B]']) -> 'Eff[B]': pass
    @overload
    def catch(self, fn: Callable[[Exc], Awaitable[B]]) -> 'Eff[B]': pass
    @overload
    def catch(self, fn: Callable[[Exc], B]) -> 'Eff[B]': pass

    def catch(self, fn):
        """
           Catch errors(Exception heirs) in all deeper nested functions.
           It can return both Eff and non-Eff entities.
           It can accept both async and sync functions. Async functions are awaited.
           MonadError is not suppressed.
        """

        async def new_effect():
            try:
                return await _maybe_await(self.effect())
            except Exception as err:
                if isinstance(err, (MonadError, KeyboardInterrupt)):
                    raise err
                current = await _maybe_await(fn(err))
                if isinstance(current, self.__class__):
                    return await _maybe_await(current.effect())
                return current
        return Eff(new_effect)

    def ensure(self, fn: Callable[[], Union[Awaitable[None], None]]) -> 'Eff[A]':
        """Guaranteed to execute the function-parameter, similar to try finally.
           It can accept both async and sync functions. Async functions are awaited.
        """

        async def new_effect():
            try:
                return await _maybe_await(self.effect())
            finally:
                await _maybe_await(fn())
        return Eff(new_effect)

    def to_task(self) -> asyncio.Task:
        """Wraps the inner effect into a Task. Inner effect must be a coroutine function.
           :raises MonadError: inner effect is a sync function
        """
        panics.on_sync(self.effect, monad_name=self.__class__.__name__, method='to_task')
        return asyncio.create_task(self.effect())

    async def run(self, delay: Optional[Union[int, float]] = None) -> A:
        """
           Async - starts the chain.
           :raises TimeoutError: delay is not None and the waiting time has been exceeded.
        """
        if delay is None:
            return await _maybe_await(self.effect())
        else:
            async with asyncio.timeout(delay=delay):
                return await _maybe_await(self.effect())

    @staticmethod
    def of(value: A) -> 'Eff[A]':
        """Wraps a non-Eff value in the container. No inspections here."""
        return Eff(lambda: value)

    def __repr__(self):
        return f"Eff({self.effect})"
